@@ -3,30 +3,30 @@ Copyright 2022 by Jiahui Huang. All rights reserved.
 This file is part of PyCG toolbox and is released under "MIT License Agreement".
 Please see the LICENSE file that should have been included as part of this package.
 """
-
 import functools
-import pdb
-import shutil
-import json
 import os
+import pdb
 import tempfile
+import uuid
+import copy
+import pickle
+import textwrap
+from pathlib import Path
+from typing import List, Dict, Tuple, Union, Optional, Callable, Any
+
 import numpy as np
+from pyquaternion import Quaternion
+
 from .animation import SceneAnimator
-from .isometry import Isometry
+from .isometry import Isometry, ScaledIsometry
 import pycg.vis as vis
 import pycg.image as image
 from pycg.exp import logger
 import pycg.o3d as o3d
 import pycg.blender_client as blender
 from pycg import get_assets_path
-from pathlib import Path
-from pyquaternion import Quaternion
-import uuid
-import copy
-import pickle
-import textwrap
-from open3d.visualization import gui
 
+gui = o3d.visualization.gui
 
 class GLEngineWrapper:
     def __init__(self, engine):
@@ -349,7 +349,7 @@ class VisualizerManager:
 
         if use_new_api:
             # Please refer to Open3D/python/open3d/visualization/draw.py
-            gui.Application.instance.initialize()
+            gui.Application.instance.initialize(str(o3d.get_resource_path()))
 
             all_windows = []
             for scene_idx, (cur_scene, cur_title) in enumerate(zip(self.scenes, self.scene_titles)):
@@ -737,7 +737,7 @@ class CameraIntrinsic:
 class SceneObject:
     USD_RAW_ATTR_NAME = "pycg_raw_attributes"
 
-    def __init__(self, geom, pose: Isometry = Isometry(), attributes: dict = None):
+    def __init__(self, geom, pose: Isometry = Isometry(), scale: float = 1.0, attributes: dict = None):
         if attributes is None:
             attributes = {}
         if isinstance(geom, vis.AnnotatedGeometry):
@@ -747,6 +747,7 @@ class SceneObject:
         else:
             self.geom, self.annotations = geom, None
         self.pose = pose
+        self.scale = scale
         self.visible = True
         """
         List of Attributes:
@@ -766,18 +767,26 @@ class SceneObject:
         """
         self.attributes = attributes
 
+    @property
+    def scaled_iso(self):
+        return ScaledIsometry.from_inner_form(self.pose, self.scale)
+
     def get_extent(self):
         # Note this is only a rough extent!
+        if 'text_pos' in self.attributes:
+            ps = np.asarray(self.attributes['text_pos'])
+            return ps, ps + 0.01
+            
         bound_points = self.geom.get_axis_aligned_bounding_box().get_box_points()
         if hasattr(bound_points, "numpy"):
             bound_points = bound_points.numpy()
         else:
             bound_points = np.asarray(bound_points)
-        bound_points = self.pose @ bound_points
+        bound_points = self.scaled_iso @ bound_points
         return np.min(bound_points, axis=0), np.max(bound_points, axis=0)
 
     def get_transformed(self):
-        return self.pose @ self.geom
+        return self.scaled_iso @ self.geom
 
     def get_filament_material(self, scene_shader, scene_point_size):
         mat = o3d.visualization.rendering.MaterialRecord()
@@ -829,6 +838,14 @@ class SceneObject:
         if "uniform_color" in self.attributes:
             mat.base_color = self.attributes["uniform_color"]
 
+        if hasattr(self.geom, "textures"):
+            if len(self.geom.textures) == 1:
+                tex_img = np.asarray(self.geom.textures[0])[::-1]
+                tex_img = np.ascontiguousarray(tex_img)
+                mat.albedo_img = o3d.geometry.Image(tex_img)
+            elif len(self.geom.textures) > 1:
+                logger.warning(f"More than one texture found for {self.geom}, texture will not be displayed.")
+
         return mat
 
     def add_usd_prim(self, stage, prim_path):
@@ -878,6 +895,14 @@ class SceneObject:
             UsdShade.MaterialBindingAPI(usd_mesh).Bind(mesh_material)
 
         elif isinstance(self.geom, o3d.geometry.PointCloud):
+            # xyz = np.asarray(self.geom.points)
+            # from tqdm import tqdm
+            # for idx, pts in enumerate(tqdm(xyz)):
+            #     # print(idx)
+            #     usd_pcd = UsdGeom.Sphere.Define(stage, prim_path + f"/s{idx}")
+            #     usd_pcd.AddTransformOp().Set(Gf.Matrix4d(Isometry(t=pts).matrix.T))
+            #     usd_pcd.GetRadiusAttr().Set(0.05)
+
             usd_pcd = UsdGeom.Points.Define(stage, prim_path)
             usd_pcd.AddTransformOp().Set(Gf.Matrix4d(self.pose.matrix.T))
             if self.geom.has_points():
@@ -992,7 +1017,6 @@ class Scene:
     def __init__(self, cam_path: str = None, up_axis: str = '+Y'):
         self.objects = {}
         self.lights = {}
-        self.output = ["rgb"]
 
         self.cached_pyrenderer = None
         self.selection_sets = []
@@ -1200,7 +1224,7 @@ class Scene:
         if isinstance(geom, SceneObject):
             new_obj = copy.deepcopy(geom)
         else:
-            new_obj = SceneObject(geom, pose, attributes)
+            new_obj = SceneObject(geom, pose=pose, attributes=attributes)
         self.objects[new_name] = new_obj
         return self if not return_name else new_name
 
@@ -1317,6 +1341,12 @@ class Scene:
             obj.pose.t -= center
         return self
 
+    def _build_raycasting_scene(self):
+        scene = o3d.t.geometry.RaycastingScene()
+        for obj in self.objects.values():
+            scene.add_triangles(obj.geom)
+        return scene
+
     def _build_gl_engine(self, window_name, visible, pos=None):
         if pos is None:
             pos = (50, 50)
@@ -1398,14 +1428,16 @@ class Scene:
         for mesh_name, mesh_obj in self.objects.items():
             mat = mesh_obj.get_filament_material(self.viewport_shading, int(self.point_size))
             if "text" in mesh_obj.attributes:
-                sv.add_3d_label(mesh_obj.pose @ mesh_obj.attributes["text_pos"], mesh_obj.attributes["text"])
+                if visualizer is not None:
+                    sv.add_3d_label(mesh_obj.pose @ mesh_obj.attributes["text_pos"], mesh_obj.attributes["text"])
             else:
                 # Origin is mesh_obj.get_transformed()
                 sv.add_geometry(mesh_name, mesh_obj.geom, mat)
-                scene.set_geometry_transform(mesh_name, mesh_obj.pose.matrix)
+                pose_matrix = mesh_obj.scaled_iso.matrix
+                scene.set_geometry_transform(mesh_name, pose_matrix)
                 scene.set_geometry_double_sided(mesh_name, not self.backface_culling)
 
-        import open3d.visualization.rendering as o3dr
+        o3dr = o3d.visualization.rendering
         scene.view.set_color_grading(o3dr.ColorGrading(o3dr.ColorGrading.Quality.ULTRA,
                                                        o3dr.ColorGrading.ToneMapping.LINEAR))
         scene.view.set_shadowing(True, scene.view.ShadowType.VSM)
@@ -1450,7 +1482,8 @@ class Scene:
             engine.setup_camera(self.camera_intrinsic.to_open3d_intrinsic(), self.camera_pose.inv().matrix)
         for obj_uuid in self.animator.events.keys():
             if scene.has_geometry(obj_uuid):
-                scene.set_geometry_transform(obj_uuid, self.objects[obj_uuid].pose.matrix)
+                pose_matrix = self.objects[obj_uuid].scaled_iso.matrix
+                scene.set_geometry_transform(obj_uuid, pose_matrix)
             if obj_uuid in self.lights.keys():
                 self.lights[obj_uuid].setup_filament_scene(obj_uuid, scene.scene, self, update=True)
 
@@ -1489,6 +1522,16 @@ class Scene:
         vis_manager.add_scene(self, title, self.record_camera_pose if allow_change_pose else None)
         vis_manager.run(use_new_api=use_new_api, key_bindings=key_bindings)
         return self
+    
+    def preview_polyscope(self):
+        import polyscope as ps
+        ps.init()
+        for obj_name, obj in self.objects.items():
+            if isinstance(obj, o3d.geometry.PointCloud):
+                ps_pcd = ps.register_point_cloud(obj_name, np.asarray(obj.points))
+                if obj.has_color():
+                    ps_pcd.add_color_quantity("color", np.asarray(obj.colors))
+        ps.show()
 
     def _setup_blender_static(self):
         # Set-up the static contents of the scene
@@ -1543,13 +1586,18 @@ class Scene:
 
         return rgb_img
 
-    def render_blender_animation(self, do_render: bool = True, quality: int = 128):
+    def render_blender_animation(self, do_render: bool = True, quality: int = 128,
+                                 start_t: Optional[int] = None, end_t: Optional[int] = None):
         self._setup_blender_static()
         self.animator.send_blender()
         if not do_render or 'PAUSE_BEFORE_RENDER' in os.environ.keys():
             blender.poll_notified()
 
         t_start, t_end = self.animator.get_range()
+        if start_t is not None:
+            t_start = start_t
+        if end_t is not None:
+            t_end = end_t
         for t_cur in range(t_start, t_end + 1):
             # This will have the same time of 'render_animation'
             blender.send_eval(f"bpy.context.scene.frame_set({t_cur})")
@@ -1560,6 +1608,33 @@ class Scene:
                 rgb_img = image.read(render_tmp_file)
             yield t_cur, rgb_img
 
+    def render_opengl_depth(self):
+        gl_engine = self._build_gl_engine("_", False)
+        engine = gl_engine.engine
+        engine.poll_events()
+        engine.update_renderer()
+        captured_depth = engine.capture_depth_float_buffer(do_render=True)
+        engine.destroy_window()
+        return np.asarray(captured_depth)
+    
+    def render_opengl_depth_animation(self):
+        gl_engine = self._build_gl_engine("_", False)
+        engine = gl_engine.engine
+
+        t_start, t_end = self.animator.get_range()
+        for t_cur in range(t_start, t_end + 1):
+            self.animator.set_frame(t_cur)
+            self._update_gl_engine(gl_engine)
+            engine.poll_events()
+            engine.update_renderer()
+            captured_depth = np.asarray(engine.capture_depth_float_buffer(do_render=True))
+            yield t_cur, captured_depth
+
+        engine.destroy_window()
+
+    def render_raycasting_depth(self):
+        raise NotImplementedError
+
     def render_opengl(self, multisample: int = 1, need_alpha=False, save_path: str = None):
         saved_intrinsic = copy.copy(self.camera_intrinsic)
         self.camera_intrinsic = self.camera_intrinsic.scale(multisample)
@@ -1747,6 +1822,41 @@ class ThemeAngela(BaseTheme):
         scene.additional_blender_commands = "bpy.data.scenes[0].view_settings.view_transform = 'Standard'"
 
 
+class ThemeNormalShadow(BaseTheme):
+    def __init__(self, smooth_shading: bool = False,
+                 sun_tilt_right: float = 0.0, sun_tilt_back: float = 0.0,
+                 sun_energy: float = 1.5, sun_angle: float = 20.0,
+                 normal_saturation: float = 1.5):
+        super().__init__('''
+        Normal rendering with shadow casted (for blender).
+        ''')
+        self.smooth_shading = smooth_shading
+        self.sun_tilt_right = sun_tilt_right
+        self.sun_tilt_back = sun_tilt_back
+        self.sun_energy = sun_energy
+        self.sun_angle = sun_angle
+        self.normal_saturation = normal_saturation  # blender default is 1.0
+
+    def apply_to(self, scene: Scene):
+        scene.lights.clear()
+        scene.film_transparent = True
+        scene.viewport_shading = 'LIT'
+        scene.remove_object('auto_plane', non_exist_ok=True)
+        scene.env_map = None
+        for o in scene.objects.values():
+            o.attributes['smooth_shading'] = self.smooth_shading
+            o.attributes['material.ao'] = {"on": False}
+            o.attributes['material.normal'] = {"on": True, "saturation": self.normal_saturation}
+        scene.auto_plane(dist_ratio=0.0)
+        scene.objects['auto_plane'].attributes['cycles.is_shadow_catcher'] = True
+        scene.ambient_color = (1.0, 1.0, 1.0, 0.6)
+
+        light_iso = self.determine_sun_iso(scene, self.sun_tilt_back, self.sun_tilt_right)
+        scene.add_light_sun(
+            light_dir=light_iso.q.q, light_energy=self.sun_energy, angle=self.sun_angle / 180.0 * np.pi)
+        scene.additional_blender_commands = "bpy.data.scenes[0].view_settings.view_transform = 'Standard'"
+
+
 class ThemeDiffuseShadow(BaseTheme):
     def __init__(self, base_color=None, smooth_shading: bool = False,
                  sun_tilt_right: float = 0.0, sun_tilt_back: float = 0.0,
